package tickets;

import java.io.*;
import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Vector;

/*
 *  class Tickets
 *  Tickets are pseudoprime BigIntegers. This class performs the following tasks:
 *  -Generate a ticket
 *  -Generate a range of tickets and store it in a tickets file
 *  -Read a single ticket from a tickets file
 *  -Read a range of tickets from a tickets file
 */
public class Tickets 
{
	public static BigInteger generateTicket(int bitlength) throws ArithmeticException
	{
		//return BigInteger.probablePrime(bitlength, new java.security.SecureRandom());
		return findStrongPrimeRecursive(bitlength, bitlength/2, 5)[0];
	}
	
	/*
     * Returns p, strong prime of "bitlength", and r, s primes (possible strong) such that r | (p-1) and s | (p+1).
	 * Method can be found in J. Gordon. "Strong primes are easy to find". I've added recursivity.
     * Recursivity stops when a "bottomTopBitlength" is reached.
     * Certainty of the primality test is (1-0.5^millerRabinIterations). 
     */
	private static BigInteger[] findStrongPrimeRecursive(int bitlength, int bottomTopBitlength, int millerRabinIterations)
	{
		BigInteger r, s, p, p0, rs, amount;
				
		if(bitlength <= bottomTopBitlength)
			return new BigInteger[] {BigInteger.probablePrime(bitlength, new SecureRandom()), BigInteger.ZERO, BigInteger.ZERO};
		
		//get r and s
		r = findStrongPrimeRecursive((bitlength/2)-10, bottomTopBitlength, millerRabinIterations)[0];
		s =findStrongPrimeRecursive((bitlength/2)-10, bottomTopBitlength, millerRabinIterations)[0];
		//calculate p0
		rs = r.multiply(s);
		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
		if(!p0.testBit(0))	//if even
			p0 = p0.add(rs);
		amount = rs.shiftLeft(1);	//multiply by 2
		p = p0;
		
		for(;;)
		{
			p = p.add(amount);	
			if(p.bitLength() < bitlength)
				continue;
			if(passesMillerRabin(p,millerRabinIterations))
				break;
		}
		
//		System.out.println("Found "+p.bitLength()+" bits prime");
//		System.out.println("\t p0: "+p0.bitLength());
//		System.out.println("\t amount: "+amount.bitLength());
//		System.out.println("\t r: "+r.bitLength());
//		System.out.println("\t s: "+s.bitLength());
		return new BigInteger[] {p, r, s};
	}
    
	/*
	 * Miller-Rabin primality test.
	 */
    private static boolean passesMillerRabin(BigInteger number, int iterations) 
    { 	
    	BigInteger TWO = BigInteger.valueOf(2);
    	// Find a and m such that m is odd and this == 1 + 2**a * m
        BigInteger thisMinusOne = number.subtract(BigInteger.ONE);
    	BigInteger m = thisMinusOne;
    	int a = m.getLowestSetBit();
    	m = m.shiftRight(a);

    	// Do the tests
            SecureRandom rnd = new SecureRandom();
    	for (int i=0; i<iterations; i++) {
    	    // Generate a uniform random on (1, this)
    	    BigInteger b;
    	    do {
    		b = new BigInteger(number.bitLength(), rnd);
    	    } while (b.compareTo(BigInteger.ONE) <= 0 || b.compareTo(number) >= 0);

    	    int j = 0;
    	    BigInteger z = b.modPow(m, number);
    	    while(!((j==0 && z.equals(BigInteger.ONE)) || z.equals(thisMinusOne))) {
    		if (j>0 && z.equals(BigInteger.ONE) || ++j==a)
    		    return false;
    		z = z.modPow(TWO, number);
    	    }
    	}
    	return true;
        }
	
	/*
	 * Returns a new BigInteger that represents a ticket, of bitlength
	 * Throws ArithmeticException if bitLength < 2
	 */
	public static BigInteger generateTicketOld(int bitlength) throws ArithmeticException
	{
			return BigInteger.probablePrime(bitlength, new java.security.SecureRandom());
	}
	
	/*
	 *  Generates numberOfTickets tickets of bitlength and stores them in filename
	 *  Returns true if OK, false otherwise.
	 *  File format is:
	 *  	numberOfTickets bitlength
	 *  	ticket0
	 *  	ticket1
	 *  	...
	 */
	public static  boolean storeNewTicketRange(String filename, int numberOfTickets, int bitlength)
	{
		PrintStream p = null;
		try
		{
			p = new PrintStream(new FileOutputStream(filename));
			//print the number of tickets
			p.print(numberOfTickets+" ");
			//print the bitlength
			p.println(bitlength);
			//generate and print the tickets
			for(int i = 0; i < numberOfTickets; i++)
			{
				p.println(generateTicket(bitlength));
				if(i%100==0)
					System.out.println(i+" out of "+numberOfTickets+" tickets created.");
			}
			p.print(generateTicket(bitlength));
			System.out.println(numberOfTickets+" tickets created.");
			return true;
		}
		catch(ArithmeticException ae)
		{
			System.out.println("Bitlength must be greater or equal than 2. File "+filename+" could not be created.");
			return false;
		}
		catch(Exception e)
		{
			System.out.println("File "+filename+" could not be created.");
			return false;
		}
		finally
		{
			p.close();
		}
	}
	
	/*
	 * Returns the ticket stored in the line Id+1 of filename. Lines start at 0 (for Id=0 returns the first line).
	 * Returns null if Id exceeds the number of lines.
	 * Throws:
	 * 	-FileNotFoundException if the file does not exist.
	 * 	-IOException if there is an error in the file, or
	 * 	-NumberFormatException if there is an error in the file
	 */
	public static BigInteger getTicketFromFile(String filename, int id) throws FileNotFoundException, IOException, NumberFormatException
	{
		BufferedReader in = null;
		try
		{
			in = new BufferedReader(new FileReader(filename));
			//skip the first line
			in.readLine();
			//go to the required line
			for(int i = 0; i < id; i++)
				in.readLine();
			BigInteger ticket = new BigInteger(in.readLine());
			in.close();
			
			return ticket;
		}
		catch(FileNotFoundException fnfe)
		{
			System.out.println("File "+filename+" not found.");
			return null;
		}
		catch(Exception e)
		{
			System.out.println("There was an error while reading "+filename+".");
			return null;
		}
	}
	
	/*
	 * Returns a range of tickets stored in filename, from startId to endId (both included; if startId==endId an element is returned). The tickets are stored in a BigInteger[].
	 * Returns an empty BigInteger[] if startId > endId
	 * Throws:
	 * 	-FileNotFoundException if the file does not exist.
	 * 	-IOException if there is an error in the file, or endId is beyond the end of the file, or
	 * 	-NumberFormatException if there is an error in the file
	 *		-NullPointerException for problems related with null references or wrong input parameters
	 */
	public static Vector<BigInteger> getRangeFromFile(String filename, int startId, int endId)
	{	
		if(startId < 0 || endId < 0 || startId > endId)
		{
			System.out.println("endId should be greater than startId and both greater than zero.");
			return null;
		}
		
		Vector<BigInteger> tickets = new Vector<BigInteger>(endId-startId);
		BufferedReader in = null;
		
		try
		{
			in = new BufferedReader(new FileReader(filename));
			//skip the first line
			in.readLine();
			//go to startId
			for(int i = 0; i < startId; i++)
				in.readLine();
			//start reading. if endId is beyond the end of the file a IOException should be thrown
			for(int i = 0; i < (endId-startId+1); i++)
				tickets.add(new BigInteger(in.readLine()));
			in.close();
			return tickets;
		}
		catch(FileNotFoundException fnfe)
		{
			System.out.println("File "+filename+" not found.");
			return null;
		}
		catch(Exception e)
		{
			System.out.println("There is an error in "+filename+", or maybe endId goes beyond the end of the file.");
			return null;
		}
	}
	
	public static Vector<BigInteger> getAllTicketsFromFile(String filename)
	{
		BufferedReader in = null;
		String[] firstLine;
		int numberOfTickets;
		
		try
		{
			in = new BufferedReader(new FileReader(filename));
			//read the number of tickets (first argument in first line)
			firstLine = in.readLine().split(" ");
			numberOfTickets = new Integer(firstLine[0]);
			//ticketsBitlength = new Integer(firstLine[1]);
			Vector<BigInteger> tickets = new Vector<BigInteger>(numberOfTickets);
			//read all tickets
			for(int i = 0; i < numberOfTickets; i++)
				tickets.add(new BigInteger(in.readLine()));
			in.close();
			
			return tickets;		
		}
		catch(FileNotFoundException fnfe)
		{
			System.out.println("File "+filename+" not found.");
			return null;
		}
		catch(Exception e)
		{
			System.out.println("There is an error in "+filename);
			return null;
		}		
	}
	
	public static void main(String[] args)
	{
//		System.out.println(generateTicket(128));
//		storeNewTicketRange("./pruebatickets.txt",10,64);
		try
		{
//			System.out.println(getTicketFromFile("./pruebatickets.txt",9));
//			BigInteger[] tickets = getRangeFromFile("./pruebatickets.txt",-1,9);
//			for(int i = 0; i < tickets.length; i++)
//				System.out.println(tickets[i]);
			
			int numberOfTickets = 0;
			int ticketsBitlength = 0;
			
			long startTime = System.nanoTime();
			storeNewTicketRange("./pruebaTickets.txt",1000,64);
			long endTime = System.nanoTime();
			System.out.println("Elapsed secs: "+(endTime-startTime)/1.0E09);			
			
			Vector<BigInteger> tickets = getAllTicketsFromFile("./pruebaTickets.txt");
			System.out.println(tickets.size() + " tickets read");
		}
		catch(Exception e)
		{
			System.out.println(e.getMessage());
		}
	}

}
