package keyserver;

import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Vector;

/*
 * Performs some needed mathematical operations with BigIntegers.
 */
public class MathOps 
{
	public static final BigInteger TWO = BigInteger.valueOf(2);
	
	/*
	 * Multiplies all the BigIntegers contained in vector, and returns the result.
	 */
	public static BigInteger productory(Vector<BigInteger> vector)
	{
		BigInteger result = BigInteger.ONE;
		
		for(int i = 0; i < vector.size(); i++)
			result = result.multiply(vector.get(i));
		
		return result;
	}
	
    /*
     * Returns [d,u,v] such that ua + vb = d
     * In the protocol, a is delta and b is L
     */
    public static BigInteger[] extendedEuclides(BigInteger a, BigInteger b)
    {
	    	BigInteger[] result = new BigInteger[3];
	    	if(b.compareTo(BigInteger.ZERO)==0)
	    	{
	    		result[0] = a; result[1] = BigInteger.ONE; result[2] = BigInteger.ZERO;
	    		return result;
	    	}
	    	else
	    	{
	    		BigInteger aux;
	    		result = extendedEuclides(b,a.mod(b));
	    		aux = result[2]; result[2] = result[1].subtract(result[2].multiply(a.divide(b))); result[1] = aux;
	    		return result;
	    	}
    }
      
    /*
     * Returns m, strong prime, p (possible strong) prime and q such that m = p*q +1
     */
    public static BigInteger[] findMPQ(int mBitlength)
    {
    	//findStrongPrimeRecursive returns m, p and other value.
    	//we want m, p and q
    	BigInteger[] result = findStrongPrimeRecursive(mBitlength, mBitlength/2, 5);
    	result[2] = result[0].subtract(BigInteger.ONE).divide(result[1]);
    	
    	return result;   	
    }
    
	/*
     * Returns p, strong prime of "bitlength", and r, s primes (possible strong) such that r | (p-1) and s | (p+1).
	 * Method can be found in J. Gordon. "Strong primes are easy to find". I've added recursivity.
     * Recursivity stops when a "bottomTopBitlength" is reached.
     * Certainty of the primality test is (1-0.5^millerRabinIterations). 
     */
	private static BigInteger[] findStrongPrimeRecursive(int bitlength, int bottomTopBitlength, int millerRabinIterations)
	{
		BigInteger r, s, p, p0, rs, amount;
				
		if(bitlength <= bottomTopBitlength)
			return new BigInteger[] {BigInteger.probablePrime(bitlength, new SecureRandom()), BigInteger.ZERO, BigInteger.ZERO};
		
		//get r and s
		r = findStrongPrimeRecursive((bitlength/2)-10, bottomTopBitlength, millerRabinIterations)[0];
		s =findStrongPrimeRecursive((bitlength/2)-10, bottomTopBitlength, millerRabinIterations)[0];
		//calculate p0
		rs = r.multiply(s);
		p0 = s.modPow(r.subtract(BigInteger.ONE), rs).subtract(r.modPow(s.subtract(BigInteger.ONE), rs)).mod(rs);
		if(!p0.testBit(0))	//if even
			p0 = p0.add(rs);
		amount = rs.shiftLeft(1);	//multiply by 2
		p = p0;
		
		for(;;)
		{
			p = p.add(amount);	
			if(p.bitLength() < bitlength)
				continue;
			if(passesMillerRabin(p,millerRabinIterations))
				break;
		}
		
//		System.out.println("Found "+p.bitLength()+" bits prime");
//		System.out.println("\t p0: "+p0.bitLength());
//		System.out.println("\t amount: "+amount.bitLength());
//		System.out.println("\t r: "+r.bitLength());
//		System.out.println("\t s: "+s.bitLength());
		return new BigInteger[] {p, r, s};
	}
    
	/*
	 * Miller-Rabin primality test.
	 */
    private static boolean passesMillerRabin(BigInteger number, int iterations) 
    { 	
    	// Find a and m such that m is odd and this == 1 + 2**a * m
        BigInteger thisMinusOne = number.subtract(BigInteger.ONE);
    	BigInteger m = thisMinusOne;
    	int a = m.getLowestSetBit();
    	m = m.shiftRight(a);

    	// Do the tests
            SecureRandom rnd = new SecureRandom();
    	for (int i=0; i<iterations; i++) {
    	    // Generate a uniform random on (1, this)
    	    BigInteger b;
    	    do {
    		b = new BigInteger(number.bitLength(), rnd);
    	    } while (b.compareTo(BigInteger.ONE) <= 0 || b.compareTo(number) >= 0);

    	    int j = 0;
    	    BigInteger z = b.modPow(m, number);
    	    while(!((j==0 && z.equals(BigInteger.ONE)) || z.equals(thisMinusOne))) {
    		if (j>0 && z.equals(BigInteger.ONE) || ++j==a)
    		    return false;
    		z = z.modPow(TWO, number);
    	    }
    	}
    	return true;
        }
    
	/*
	 * Returns {k, delta} such that delta = k + p, with delta smaller than every ticket used
	 * It is clear that p should be smaller than delta
	 */
	public static BigInteger[] findKDelta(BigInteger p, int deltaBitlength)
	{
		BigInteger k;
		BigInteger delta;
		
		do
		{
			delta = new BigInteger(deltaBitlength, new SecureRandom());
			k = delta.subtract(p);
		}while(k.compareTo(BigInteger.ZERO)<1);	//if k<=0 then delta <= p, that's not acceptable
					
		return new BigInteger[] {k, delta};
	}
	
	/*
	 * Returns a primitive element of Zm
	 * Lemma: Suppose m is prime, and the factorization of m-1 is "m-1 = m1^e1*...mz^ez". Then a in Zm* is a primitive element if and only if
	 * a^{(m-1)/mj} != 1 mod m for 1<= j <=z
	 */
	public static BigInteger findPrimitiveElement(BigInteger m, BigInteger[] mLess1Factors)
	{
		SecureRandom sr = new SecureRandom();
		BigInteger a;
		BigInteger mLess1 = m.subtract(BigInteger.ONE);
		BigInteger exponent;
		BigInteger power;
		boolean success;
		
		do
		{
			success = true;
			
			a  = new BigInteger(m.bitLength(),sr); //random value
			if(a.compareTo(m)>0)
				a = a.mod(m);
			for(int i = 0; i < mLess1Factors.length; i++)
			{
				exponent = mLess1.divide(mLess1Factors[i]);
				power = a.modPow(exponent,m);
				if(power.compareTo(BigInteger.ONE)==0) //must be != 1
				{
					success = false;
					break;
				}
			}
		}while(!success);
		
		return a;
	}

	/*
	 * Returns g such that g^p=1 mod m
	 */
	public static BigInteger findGenerator(BigInteger m, BigInteger p, BigInteger q)
	{
		//a is a primitive element of Zm
		BigInteger a = findPrimitiveElement(m,new BigInteger[]{p,q});
		
		return a.modPow(q, m);
	}
	
	/*
	 * Returns the secret r = g^k mod m
	 */
	public static BigInteger calculateSecret(BigInteger g, BigInteger k, BigInteger m)
	{
		return g.modPow(k, m);
	}	
}
